# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection and extraction

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Fri Jul 12 14:05:16 2013
"""

def filter_times(times, min, max):
    return tuple(t for t in times if (t > min) and (t < max))


def realign_spikes(times, spikes, upsample, jitter=1, reflect_fft=False):
    """Realign spikes to their peaks using bandwidth-limited resampling

    times    : one-dimensional array of spike times, in units of samples
    spikes   : array of spike waveforms, with dimensions (nspikes, npoints)
    upsample : integer, the factor by which to upsample the data for realignment
    jitter   : samples (in original data) to search for true peak

    Returns (times, spikes), with the sampling rate increased by a factor of upsample

    """
    from numpy import mean, zeros, asarray
    upsample = int(upsample)
    assert upsample > 1, "Upsampling factor must be greater than 1"
    nevents, nsamples = spikes.shape

    # first infer the expected peak time
    expected_peak = mean(spikes, 0).argmax() * upsample
    spikes = fftresample(spikes, nsamples * upsample, reflect=reflect_fft)
    # find peaks within upsample samples of mean peak
    shift = find_peaks(spikes, expected_peak, upsample * jitter)
    start = shift + upsample * jitter
    nshifted = (nsamples - 2 * jitter) * upsample
    shifted = zeros((nevents, nshifted))
    for i,spike in enumerate(spikes):
        shifted[i,:] = spike[start[i]:start[i]+nshifted]
    return (asarray(times) * upsample + start, shifted)


def find_peaks(spikes, peak, window):
    """Locate the peaks in an array of spikes.

    spikes: resampled spike waveforms, dimensions (nspikes, nsamples)
    peak:   the expected peak location
    window: the number of samples to either side of the peak to look for the peak

    Returns array of shift values relative to peak

    """
    r = slice(peak - window, peak + window + 1)
    return spikes[:,r].argmax(1) - window


def fftresample(S, npoints, axis=1, reflect=False):
    """Resample a signal using discrete fourier transform. The signal
    is transformed in the fourier domain and then padded or truncated
    to the correct sampling frequency.  This should be equivalent to
    a sinc resampling.

    Set reflect to True to pad the sample with reflected copies on either end.
    Do this if the shape of the spike really matters.

    """
    from numpy import column_stack
    from numpy.fft import rfft, irfft
    if reflect:
        Srev = S[:,::-1]
        S = column_stack([Srev, S, Srev])
        npoints *= 3
    Sf = rfft(S, axis=axis)
    Srs = (1. * npoints / S.shape[axis]) * irfft(Sf, npoints, axis=axis)
    if reflect:
        npoints /= 3
        return Srs[:,npoints:npoints*2]
    else:
        return Srs



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection routines

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Wed Jul 24 09:26:36 2013
"""

__version__ = "1.2.0"
from .spikes import detector, peaks, subthreshold
from .tools import realign_spikes, filter_times



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""
Functions to extract spikes from an intracellular recording

Copyright (C) 2012 Dan Meliza <dmeliza@gmail.com>
Created 2012-03-21
"""

import numpy as nx


def spike_times(x, spike_thresh=5, dt=0.02, peak_window=1.5, **kwargs):
    """ Detect spikes in a signal

    Returns an array of times corresponding to the peaks in the signal that were
    above spike_thresh.

    - x : signal to analyze
    - spike_thresh: the threshold for the detector
    - dt : time step of signal
    - peak_window: how far to search forward (in dt units) for the peak
    """
    from quickspikes._spikes import spike_times as st
    t = st(x, spike_thresh, int(peak_window/dt))
    return t.nonzero()[0]

def spike_waveforms(x, tt, dt=0.02, peak_window=1.5, spike_onset=3.5, spike_offset=8.,
                    min_isi=8., **kwargs):
    """ Extract spike waveforms

    Returns a 2D array containing the waveforms of the spikes in x with peaks at tt

    - x : signal to analyze
    - tt : times of the peaks (in samples)
    - dt : time step of signal
    - spike_onset : amount of time (in dt units) before peak to include
    - spike_offset : amount of time (in dt units) after peak to include
    - min_isi : exclude spikes with isi < min_isi
    """
    from quickspikes._spikes import extract_spikes
    # drop spikes that won't give a complete waveform for this step
    tta = tt[(tt > int(spike_onset/dt)) & ((tt + int(spike_offset/dt)) < x.size)]
    # drop spikes with isi < min_isi
    ind = ((nx.diff(tta) * dt) > min_isi).nonzero()[0]
    tta = tta[ind+1]
    if tta.size > 0:
        return extract_spikes(x, tta, int(spike_onset/dt), int(spike_offset/dt)).T
    else:
        return None

def extract_subthreshold(x, tt, dt=0.02, thresh_v=-50, thresh_dv=0, min_size=10, **kwargs):
    """ Extract subthreshold activity

    Spikes are removed from the voltage trace by beginning at each peak and
    moving in either direction until V drops below thresh_v OR dv drops below
    thresh_dv.

    - x : signal to analyze
    - tt : times of the peaks (in samples)
    - dt : time step of signal
    - thresh_v : do not include points contiguous with the peak with V > thresh_v
    - thresh_dv : do not include points contiguous with the peak with dV/dt > thresh_dv.
                  negative values corrspond to positive-going voltages
    - min_size : always remove at least this many points on either side of peak
    """
    from quickspikes._spikes import extract_subthreshold as es
    return es(x, tt, thresh_v, thresh_dv * dt, min_size)


def spike_lag(spk, lag=3):
    """
    Given a 2D array of spikes (T by N), return V(t), V(t-lag)
    """
    return spk[:-lag,:], spk[lag:,:]

def spike_dt(spk, lag=3):
    """
    Given a 2D array of spikes (T by N), calculate discrete difference
    for a phase space plot.  Resamples as needed.

    @return V(t), V(t)-V(t-lag)
    """
    Vt,Vtt = spike_lag(spk, lag)
    return Vt,Vtt-Vt

def fftresample(S, npoints, axis=1, padding='flip'):
    """
    Resample a signal using discrete fourier transform. The signal
    is transformed in the fourier domain and then padded or truncated
    to the correct sampling frequency.  This should be equivalent to
    a sinc resampling.
    """
    from numpy.fft import rfft, irfft
    if padding == 'flip':
        S = nx.concatenate([S[::-1], S, S[::-1]],0)
        npoints *=3
    Sf = rfft(S, axis=axis)
    S = (1. * npoints / S.shape[axis]) * irfft(Sf, npoints, axis=axis)
    if padding == 'flip':
        return S[npoints/3:npoints*2/3]
    else:
        return S


# Variables:
# End:



# -*- coding: utf-8 -*-
# -*- mode: python -*-

from nose.tools import *
import numpy as nx

# a nice surrogate spike with 20 samples before peak and 40 after
a_spike = nx.array([-1290,  -483,  -136,  -148,  -186,   637,   328,    41,    63,
                    42,   377,   872,   639,   -17,   538,   631,   530,   693,
                    743,  3456,  6345,  5868,  4543,  3087,  1691,   830,   241,
                    -350,  -567,  -996,  -877, -1771, -1659, -1968, -2013, -2290,
                    -2143, -1715, -1526, -1108,  -500,   333,    25,  -388,  -368,
                    -435,  -817,  -858,  -793, -1089,   -16,  -430,  -529,  -252,
                    -3,  -786,   -47,  -266,  -963,  -365], dtype=nx.int16)
t_peak = a_spike.argmax()
t_trough = a_spike.argmin()
a_recording = nx.zeros(10000, dtype='d')
a_times = [100, 400, 1200, 1500, 5000, 5200, 6123, 9730]
for t in a_times:
    a_recording[t:t + a_spike.size] += a_spike

b_recording = nx.load("test/intra_spike.npy").astype('d')
b_times = [7635, 8412, 9363, 10424, 11447, 12661, 13887, 15079, 16373,
           17753, 19168, 20682, 22357, 23979, 25574, 27209, 28989,
           30508, 32088, 33778]

def test_detect_extrac_spikes():
    from quickspikes.spikes import detector

    det = detector(2000, 40)
    assert_sequence_equal(det.send(a_recording), [t + t_peak for t in a_times])
    assert_sequence_equal(det.send(-a_recording), [t + t_trough for t in a_times])

@raises(ValueError)
def test_extract_spikes_nofilter():
    from quickspikes.spikes import peaks

    x = peaks(a_recording, [t + t_peak for t in a_times], n_before=20, n_after=300)

def test_extract_spikes():
    from quickspikes import peaks, filter_times

    n_before = 20
    n_after = 300
    times = filter_times([t + t_peak for t in a_times], n_before, a_recording.size - n_after)
    x = peaks(a_recording, times, n_before=n_before, n_after=n_after)
    # last peak should get dropped
    assert_equal(x.shape[0], len(a_times) - 1)
    assert_equal(x.shape[1], 320)

    assert_true(nx.all(a_spike == x[0,:a_spike.size]))

def test_detect_intrac_spikes():
    from quickspikes.spikes import detector

    det = detector(0, 100)
    assert_sequence_equal(det.send(b_recording), b_times)

    det = detector(-20, 100)
    assert_sequence_equal(det.send(b_recording), b_times)

def test_align_spikes():
    from quickspikes.spikes import peaks
    from quickspikes.tools import realign_spikes

    spikes = peaks(b_recording, b_times, 200, 400)
    times, aligned = realign_spikes(b_times, spikes, 3, 4)
    peaks = aligned.argmax(1)
    assert_true((peaks == peaks[0]).all())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- mode: python -*-
from setuptools import setup, find_packages, Extension

import numpy
try:
    numpy_include = numpy.get_include()
except AttributeError:
    numpy_include = numpy.get_numpy_include()

try:
    from Cython.Distutils import build_ext
    SUFFIX = '.pyx'
except ImportError:
    from distutils.command.build_ext import build_ext
    SUFFIX = '.c'

compiler_settings = {
    'include_dirs' : [numpy_include]
    }
_spikes = Extension('quickspikes.spikes', sources=['quickspikes/spikes' + SUFFIX],
                    **compiler_settings)

setup(
    name = 'quickspikes',
    version = '1.2.0',
    packages=find_packages(exclude=["*test*"]),
    ext_modules = [_spikes],
    cmdclass = {'build_ext': build_ext},

    install_requires = ["numpy>=1.9"],
    scripts = [],

    description = "detect and extract spikes in time series data",
    long_description = "detect and extract spikes in time series data",

    author = "Dan Meliza",
    author_email = "dan AT the domain 'meliza.org'",
    maintainer = "Dan Meliza",
    maintainer_email = "dan AT the domain 'meliza.org'",
    url= 'http://melizalab.github.com/quickspikes',
    download_url= 'https://github.com/downloads/melizalab/quickspikes',
    test_suite = 'nose.collector',
    )
